@@ -8,7 +8,8 @@ use std::{fmt::Display, ops::Index, path::PathBuf};
 
 use color_eyre::Result;
 use pilka_types::{
-    dispatch_optimal_size, ContiniousHashMap, Frame, ImageDimentions, ShaderCreateInfo,
+    dispatch_optimal_size, ContiniousHashMap, Frame, ImageDimentions, PushConstant,
+    ShaderCreateInfo, Uniform,
 };
 use pollster::block_on;
 use raw_window_handle::HasRawWindowHandle;
@@ -26,21 +27,18 @@ pub enum Pipeline {
     Compute(ComputePipeline),
 }
 
-trait Descriptor<'a, const N: usize> {
-    const DESC: [BindGroupLayoutDescriptor<'a>; N];
-
-    fn binding_group(device: &Device) -> [BindGroupLayout; N] {
-        Self::DESC.map(|x| device.create_bind_group_layout(&x))
-    }
+enum Binding {
+    Uniform,
+    Texture,
+    Sampler,
 }
 
 #[derive(Debug)]
-struct RenderPipelineLayoutInfo;
-impl<'a> RenderPipelineLayoutInfo {
-    const N: usize = 2;
-    const DESC: [BindGroupLayoutDescriptor<'a>; Self::N] = [
+struct RenderPipelineLayoutInfo<const N: usize>([BindGroupLayout; N]);
+impl<'a> RenderPipelineLayoutInfo<3> {
+    const DESC: [BindGroupLayoutDescriptor<'a>; 3] = [
         wgpu::BindGroupLayoutDescriptor {
-            label: Some("Render Bind Group Layout"),
+            label: Some("Texture Render Bind Group Layout"),
             entries: &[
                 wgpu::BindGroupLayoutEntry {
                     binding: 0,
@@ -95,17 +93,27 @@ impl<'a> RenderPipelineLayoutInfo {
             ],
         },
         wgpu::BindGroupLayoutDescriptor {
-            label: Some("Render Bind Group Layout"),
+            label: Some("Sampler Render Bind Group Layout"),
             entries: &[wgpu::BindGroupLayoutEntry {
                 binding: 0,
                 visibility: wgpu::ShaderStages::FRAGMENT,
-                ty: wgpu::BindingType::Sampler {
-                    filtering: true,
-                    comparison: false,
+                ty: wgpu::BindingType::Sampler(wgpu::SamplerBindingType::Filtering),
+                count: None,
+            }],
+        },
+        wgpu::BindGroupLayoutDescriptor {
+            label: Some("Uniform Render Bind Group Layout"),
+            entries: &[wgpu::BindGroupLayoutEntry {
+                binding: 0,
+                visibility: wgpu::ShaderStages::VERTEX_FRAGMENT,
+                ty: wgpu::BindingType::Buffer {
+                    ty: wgpu::BufferBindingType::Uniform,
+                    has_dynamic_offset: false,
+                    min_binding_size: None,
                 },
                 count: None,
             }],
-        }
+        },
         // wgpu::BindGroupLayoutDescriptor {
         //     label: Some("Fft Texture Bind Group Layout"),
         //     entries: &[wgpu::BindGroupLayoutEntry {
@@ -120,19 +128,30 @@ impl<'a> RenderPipelineLayoutInfo {
         //     }],
         // },
     ];
+
+    pub fn new(device: &Device) -> Self {
+        Self(Self::DESC.map(|x| device.create_bind_group_layout(&x)))
+    }
 }
 
-impl<'a> Descriptor<'a, { Self::N }> for RenderPipelineLayoutInfo {
-    const DESC: [BindGroupLayoutDescriptor<'a>; Self::N] = Self::DESC;
+impl<const N: usize> Index<Binding> for RenderPipelineLayoutInfo<N> {
+    type Output = BindGroupLayout;
+
+    fn index(&self, index: Binding) -> &Self::Output {
+        match index {
+            Binding::Texture => &self.0[0],
+            Binding::Sampler => &self.0[1],
+            Binding::Uniform => &self.0[2],
+        }
+    }
 }
 
 #[derive(Debug)]
-struct ComputePipelineLayoutInfo;
-impl<'a> ComputePipelineLayoutInfo {
-    const N: usize = 1;
-    const DESC: [BindGroupLayoutDescriptor<'a>; Self::N] = [
+struct ComputePipelineLayoutInfo<const N: usize>([BindGroupLayout; N]);
+impl<'a> ComputePipelineLayoutInfo<2> {
+    const DESC: [BindGroupLayoutDescriptor<'a>; 2] = [
         wgpu::BindGroupLayoutDescriptor {
-            label: Some("Compute Bind Group Layout"),
+            label: Some("Texture Compute Bind Group Layout"),
             entries: &[
                 wgpu::BindGroupLayoutEntry {
                     binding: 0,
@@ -186,6 +205,19 @@ impl<'a> ComputePipelineLayoutInfo {
                 },
             ],
         },
+        wgpu::BindGroupLayoutDescriptor {
+            label: Some("Uniform Compute Bind Group Layout"),
+            entries: &[wgpu::BindGroupLayoutEntry {
+                binding: 0,
+                visibility: wgpu::ShaderStages::COMPUTE,
+                ty: wgpu::BindingType::Buffer {
+                    ty: wgpu::BufferBindingType::Uniform,
+                    has_dynamic_offset: false,
+                    min_binding_size: None,
+                },
+                count: None,
+            }],
+        },
         // wgpu::BindGroupLayoutDescriptor {
         //     label: Some("fft Texture Bind Group Layout"),
         //     entries: &[wgpu::BindGroupLayoutEntry {
@@ -200,31 +232,36 @@ impl<'a> ComputePipelineLayoutInfo {
         //     }],
         // },
     ];
+
+    pub fn new(device: &Device) -> Self {
+        Self(Self::DESC.map(|x| device.create_bind_group_layout(&x)))
+    }
 }
 
-impl<'a> Descriptor<'a, { ComputePipelineLayoutInfo::N }> for ComputePipelineLayoutInfo {
-    const DESC: [BindGroupLayoutDescriptor<'a>; ComputePipelineLayoutInfo::N] = Self::DESC;
-}
-
-enum Binding {
-    Texture,
-    Sampler,
-    #[allow(dead_code)]
-    Fft,
-}
-
-impl<const N: usize> Index<Binding> for [BindGroupLayout; N] {
+impl<const N: usize> Index<Binding> for ComputePipelineLayoutInfo<N> {
     type Output = BindGroupLayout;
 
     fn index(&self, index: Binding) -> &Self::Output {
         match index {
-            Binding::Texture => &self[0],
-            Binding::Sampler => &self[1],
-            Binding::Fft => &self[2],
+            Binding::Texture => &self.0[0],
+            Binding::Uniform => &self.0[1],
+            _ => panic!(),
         }
     }
 }
 
+// impl<const N: usize> Index<Binding> for [BindGroupLayout; N] {
+//     type Output = BindGroupLayout;
+
+//     fn index(&self, index: Binding) -> &Self::Output {
+//         match index {
+//             Binding::Uniform => &self[0],
+//             Binding::Texture => &self[1],
+//             Binding::Sampler => &self[2],
+//         }
+//     }
+// }
+
 fn create_textures(
     device: &Device,
     extent: wgpu::Extent3d,
@@ -272,13 +309,13 @@ fn create_textures(
 
     let sampled_texture_bind_group = device.create_bind_group(&wgpu::BindGroupDescriptor {
         label: Some("Render Bind Group"),
-        layout: &RenderPipelineLayoutInfo::binding_group(device)[Binding::Texture],
+        layout: &RenderPipelineLayoutInfo::new(device)[Binding::Texture],
         entries: &entries,
     });
 
     let storage_texture_bind_group = device.create_bind_group(&wgpu::BindGroupDescriptor {
         label: Some("Compute Bind Group"),
-        layout: &ComputePipelineLayoutInfo::binding_group(device)[Binding::Texture],
+        layout: &ComputePipelineLayoutInfo::new(device)[Binding::Texture],
         entries: &entries,
     });
 
@@ -317,6 +354,11 @@ pub struct WgpuRender {
     smaa_target: smaa::SmaaTarget,
 
     pub paused: bool,
+
+    uniform: Uniform,
+    uniform_buffer: wgpu::Buffer,
+    compute_uniform_bind_group: wgpu::BindGroup,
+    render_uniform_bind_group: wgpu::BindGroup,
 }
 
 #[derive(Debug)]
@@ -443,7 +485,7 @@ impl WgpuRender {
         });
         let sampler_bind_group = device.create_bind_group(&wgpu::BindGroupDescriptor {
             label: Some("Render Bind Group"),
-            layout: &RenderPipelineLayoutInfo::binding_group(&device)[Binding::Sampler],
+            layout: &RenderPipelineLayoutInfo::new(&device)[Binding::Sampler],
             entries: &[wgpu::BindGroupEntry {
                 binding: 0,
                 resource: wgpu::BindingResource::Sampler(&sampler),
@@ -465,6 +507,31 @@ impl WgpuRender {
             smaa::SmaaMode::Smaa1X,
         );
 
+        let uniform = Uniform::default();
+        let uniform_buffer = device.create_buffer(&wgpu::BufferDescriptor {
+            label: Some("Uniform Buffer"),
+            size: std::mem::size_of::<Uniform>() as _,
+            usage: wgpu::BufferUsages::UNIFORM | wgpu::BufferUsages::COPY_DST,
+            mapped_at_creation: false,
+        });
+
+        let compute_uniform_bind_group = device.create_bind_group(&wgpu::BindGroupDescriptor {
+            label: Some("Uniform Bind Group"),
+            layout: &ComputePipelineLayoutInfo::new(&device)[Binding::Uniform],
+            entries: &[wgpu::BindGroupEntry {
+                binding: 0,
+                resource: uniform_buffer.as_entire_binding(),
+            }],
+        });
+        let render_uniform_bind_group = device.create_bind_group(&wgpu::BindGroupDescriptor {
+            label: Some("Uniform Bind Group"),
+            layout: &RenderPipelineLayoutInfo::new(&device)[Binding::Uniform],
+            entries: &[wgpu::BindGroupEntry {
+                binding: 0,
+                resource: uniform_buffer.as_entire_binding(),
+            }],
+        });
+
         Ok(Self {
             adapter,
             device,
@@ -492,6 +559,11 @@ impl WgpuRender {
             smaa_target,
 
             paused: false,
+
+            uniform,
+            uniform_buffer,
+            compute_uniform_bind_group,
+            render_uniform_bind_group,
         })
     }
 
@@ -590,8 +662,7 @@ impl WgpuRender {
             .device
             .create_pipeline_layout(&wgpu::PipelineLayoutDescriptor {
                 label: Some("Compute Pipeline Layout"),
-                bind_group_layouts: &ComputePipelineLayoutInfo::binding_group(&self.device)
-                    .each_ref(),
+                bind_group_layouts: &ComputePipelineLayoutInfo::new(&self.device).0.each_ref(),
                 push_constant_ranges: &[wgpu::PushConstantRange {
                     stages: wgpu::ShaderStages::COMPUTE,
                     range: 0..self.push_constant_ranges,
@@ -634,8 +705,7 @@ impl WgpuRender {
             .device
             .create_pipeline_layout(&wgpu::PipelineLayoutDescriptor {
                 label: Some("Render Pipeline Layout"),
-                bind_group_layouts: &RenderPipelineLayoutInfo::binding_group(&self.device)
-                    .each_ref(),
+                bind_group_layouts: &RenderPipelineLayoutInfo::new(&self.device).0.each_ref(),
                 push_constant_ranges: &[wgpu::PushConstantRange {
                     stages: wgpu::ShaderStages::VERTEX_FRAGMENT,
                     range: 0..self.push_constant_ranges,
@@ -657,9 +727,7 @@ impl WgpuRender {
                     strip_index_format: None,
                     front_face: wgpu::FrontFace::Ccw,
                     cull_mode: Some(wgpu::Face::Back),
-                    clamp_depth: false,
-                    polygon_mode: wgpu::PolygonMode::Fill,
-                    conservative: false,
+                    ..Default::default()
                 },
                 depth_stencil: None,
                 multisample: wgpu::MultisampleState {
@@ -676,14 +744,19 @@ impl WgpuRender {
                         write_mask: wgpu::ColorWrites::ALL,
                     }],
                 }),
+                multiview: None,
             });
 
         Ok(Pipeline::Render(pipeline))
     }
 
-    pub fn render(&mut self, push_constant: &[u8]) -> Result<(), wgpu::SurfaceError> {
+    pub fn render(&mut self, push_constant: PushConstant) -> Result<(), wgpu::SurfaceError> {
         puffin::profile_function!();
 
+        self.uniform = push_constant.into();
+        self.queue
+            .write_buffer(&self.uniform_buffer, 0, bytemuck::bytes_of(&self.uniform));
+
         let mut _pre_scope =
             puffin::ProfilerScope::new("Aquiring frame", puffin::short_file_name(file!()), "");
 
@@ -733,10 +806,11 @@ impl WgpuRender {
                     render_pass.set_push_constants(
                         wgpu::ShaderStages::VERTEX_FRAGMENT,
                         0,
-                        push_constant,
+                        bytemuck::bytes_of(&push_constant),
                     );
                     render_pass.set_bind_group(0, &self.sampled_texture_bind_group, &[]);
                     render_pass.set_bind_group(1, &self.sampler_bind_group, &[]);
+                    render_pass.set_bind_group(2, &self.render_uniform_bind_group, &[]);
                     render_pass.draw(0..3, 0..1);
                 }
                 Pipeline::Compute(ref pipeline) if !self.paused => {
@@ -747,8 +821,9 @@ impl WgpuRender {
                             label: Some(&format!("Compute Pass {}", i)),
                         });
                     compute_pass.set_pipeline(pipeline);
-                    compute_pass.set_push_constants(0, push_constant);
+                    compute_pass.set_push_constants(0, bytemuck::bytes_of(&push_constant));
                     compute_pass.set_bind_group(0, &self.storage_texture_bind_group, &[]);
+                    compute_pass.set_bind_group(1, &self.compute_uniform_bind_group, &[]);
                     compute_pass.dispatch(
                         dispatch_optimal_size(self.extent.width, SUBGROUP_SIZE),
                         dispatch_optimal_size(self.extent.height, SUBGROUP_SIZE),
